###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = 'Donovan Parks'
__copyright__ = 'Copyright 2014'
__credits__ = ['Donovan Parks']
__license__ = 'GPL3'
__maintainer__ = 'Donovan Parks'
__email__ = 'donovan.parks@gmail.com'

import logging
import ntpath
import os

from .exceptions import BioLibFileNotFound, BioLibDirNotFound, BioLibIOException


def canonical_gid(gid):
    """Get canonical form of NCBI genome accession.
    
    Example:
        G005435135 -> G005435135
        GCF_005435135.1 -> G005435135
        GCF_005435135.1_ASM543513v1_genomic -> G005435135
        RS_GCF_005435135.1 -> G005435135
        GB_GCA_005435135.1 -> G005435135
    """

    if gid.startswith('U'):
        return gid

    gid = gid.replace('RS_', '').replace('GB_', '')
    gid = gid.replace('GCA_', 'G').replace('GCF_', 'G')
    if '.' in gid:
        gid = gid[0:gid.find('.')]

    return gid


def is_float(s):
    """Check if a string can be converted to a float.

    Parameters
    ----------
    s : str
        String to evaluate.

    Returns
    -------
    boolean
        True if string can be converted, else False.
    """
    try:
        float(s)
    except ValueError:
        return False

    return True


def check_file_exists(input_file):
    """Assert that a file exists.

    Parameters
    ----------
    input_file : str
        The path to file being checked.

    Returns
    -------
    bool
        True if the directory exists.

    Raises
    ------
    BioLibFileNotFound
        If the file doesn't exist.
    """
    if not os.path.exists(input_file) or not os.path.isfile(input_file):
        logger = logging.getLogger('timestamp')
        logger.error('Input file does not exist: ' + input_file)
        raise BioLibFileNotFound('Input file does not exist: ' + input_file)
    return True


def check_dir_exists(input_dir):
    """Assert that a directory exists.

    Parameters
    ----------
    input_dir : str
        The path to the directory being checked.

    Returns
    -------
    bool
        True if the directory exists.

    Raises
    ------
    BioLibDirNotFound
        If the directory doesn't exist.
    """
    if not os.path.exists(input_dir) or not os.path.isdir(input_dir):
        logger = logging.getLogger('timestamp')
        logger.error('Input directory does not exist: ' + input_dir)
        raise BioLibDirNotFound('Input directory does not exist: ' + input_dir)
    return True


def make_sure_path_exists(path):
    """Create directory if it does not exist.

    Parameters
    ----------
    path : str
        The path to the directory which should be created.

    Returns
    -------
    bool
        True if the path exists.

    Raises
    ------
    BioLibIOException
        If an error was encountered while creating the directory.
    """
    if not path:
        # lack of a path qualifier is acceptable as this
        # simply specifies the current directory
        return True
    elif os.path.isdir(path):
        return True

    try:
        os.makedirs(path)
        return True
    except OSError:
        logger = logging.getLogger('timestamp')
        logger.error('Specified path could not be created: ' + path)
        raise BioLibIOException('Specified path could not be created: ' + path)


def remove_extension(filename, extension=None):
    """Remove extension from filename. A specific extension can be specified,
    otherwise the extension is taken as all characters after the last period.

    Parameters
    ----------
    filename : str
        The name of the file in which the extension will be removed.
    extension : str, optional
        The extension to be removed.

    Returns
    -------
    str
        The name of the file with the extension removed.
    """
    f = ntpath.basename(filename)

    if extension and f.endswith(extension):
        f = f[0:f.rfind(extension)]
    else:
        f = os.path.splitext(f)[0]

    if f[-1] == '.':
        f = f[0:-1]

    return f



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


import logging
import multiprocessing as mp
import os
import shutil
import subprocess
import tempfile

from tqdm import tqdm

from gtdbtk.biolib_lite.execute import check_dependencies
from gtdbtk.exceptions import GTDBTkException
from gtdbtk.io.marker.copy_number import CopyNumberFileAR122, CopyNumberFileBAC120
from gtdbtk.io.marker.tophit import TopHitPfamFile, TopHitTigrFile


class HmmAligner(object):
    """Runs HMMalign over a set of genomes."""

    def __init__(self,
                 threads,
                 pfam_top_hit_suffix,
                 tigrfam_top_hit_suffix,
                 protein_file_suffix,
                 pfam_hmm_dir,
                 tigrfam_hmm_dir,
                 bac120_markers,
                 ar122_markers):
        """Initialization."""

        check_dependencies(['hmmalign'])

        self.logger = logging.getLogger('timestamp')

        self.threads = threads
        self.pfam_top_hit_suffix = pfam_top_hit_suffix
        self.tigrfam_top_hit_suffix = tigrfam_top_hit_suffix
        self.protein_file_suffix = protein_file_suffix
        self.pfam_hmm_dir = pfam_hmm_dir
        self.tigrfam_hmm_dir = tigrfam_hmm_dir

        self.bac120_markers = bac120_markers
        self.ar122_markers = ar122_markers

        self.version = self.get_version()

    def get_version(self):
        """ get HMMER version."""
        try:
            env = os.environ.copy()
            proc = subprocess.Popen(['hmmalign', '-h'], stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, env=env, encoding='utf-8')

            output, error = proc.communicate()
            for line in output.split('\n'):
                if line.startswith('# HMMER'):
                    version = line.split(';')[0].replace('# HMMER', '').strip()
                    return version
            return "(version unavailable)"
        except:
            return "(version unavailable)"

    def align_marker_set(self, db_genome_ids, marker_set_id):
        """Threaded alignment using hmmalign for a given set of genomes.

        Parameters
        ----------
        db_genome_ids : dict
            A dictionary containing the genome ids and aa paths to process.
        marker_set_id : str
            The marker set of these genomes (bac120/ar122).

        Returns
        -------
        dict
            A dictionary of genome_ids -> aligned sequence.
        """
        q_worker = mp.Queue()
        q_writer = mp.Queue()

        for gid, gid_dict in db_genome_ids.items():
            q_worker.put((gid, gid_dict.get('aa_gene_path'), marker_set_id))
        [q_worker.put(None) for _ in range(self.threads)]

        manager = mp.Manager()
        out_dict = manager.dict()

        p_workers = [mp.Process(target=self._worker,
                                args=(q_worker, q_writer, out_dict))
                     for _ in range(self.threads)]

        p_writer = mp.Process(target=self._writer,
                              args=(q_writer, len(db_genome_ids)))

        try:
            p_writer.start()
            for p_worker in p_workers:
                p_worker.start()

            for p_worker in p_workers:
                p_worker.join()

                # Gracefully terminate the program.
                if p_worker.exitcode != 0:
                    raise GTDBTkException(
                        'hmmalign returned a non-zero exit code.')

            q_writer.put(None)
            p_writer.join()

        except Exception:
            for p in p_workers:
                p.terminate()

            p_writer.terminate()
            raise

        return {k: v for k, v in out_dict.items()}

    def _worker(self, q_worker, q_writer, out_dict):
        """The worker function, invoked in a process.

        Parameters
        ----------
        q_worker : multiprocessing.Queue
            A queue of tuples containing the genome id, aa path, and marker.
        q_writer : multiprocessing.Queue
            A queue consumed by the writer, to track progress.
        out_dict : dict
            A thread-safe managed dictionary of results.
        """
        job = q_worker.get(block=True, timeout=None)
        while job is not None:
            db_genome_id, aa_gene_path, marker_set_id = job
            out_dict[db_genome_id] = self._run_multi_align(db_genome_id,
                                                           aa_gene_path,
                                                           marker_set_id)
            q_writer.put(db_genome_id)
            job = q_worker.get(block=True, timeout=None)
        return True

    def _writer(self, q_writer, n_genomes):
        """The writer function, which reports the progress of the workers.

        Parameters
        ----------
        q_writer : multiprocessing.Queue
            A queue of genome ids which have been processed.
        n_genomes : int
            The total number of genomes to be processed.
        """
        bar_fmt = '==> Aligned {n_fmt}/{total_fmt} ({percentage:.0f}%) ' \
                  'genomes [{rate_fmt}, ETA {remaining}]'
        with tqdm(total=n_genomes, bar_format=bar_fmt) as p_bar:
            for _ in iter(q_writer.get, None):
                p_bar.update()

    def _run_multi_align(self, db_genome_id, path, marker_set_id):
        """
        Returns the concatenated marker sequence for a specific genome
        :param db_genome_id: Selected genome
        :param path: Path to the genomic fasta file for the genome
        :param marker_set_id: Unique ID of marker set to use for alignment
        """

        cur_marker_dir = os.path.dirname(os.path.dirname(path))
        pfam_tophit_file = TopHitPfamFile(cur_marker_dir, db_genome_id)
        tigr_tophit_file = TopHitTigrFile(cur_marker_dir, db_genome_id)
        pfam_tophit_file.read()
        tigr_tophit_file.read()

        if marker_set_id == 'bac120':
            copy_number_file = CopyNumberFileBAC120('/dev/null', None)
        elif marker_set_id == 'ar122':
            copy_number_file = CopyNumberFileAR122('/dev/null', None)
        else:
            raise GTDBTkException('Unknown marker set.')

        copy_number_file.add_genome(db_genome_id, path, pfam_tophit_file, tigr_tophit_file)
        single_copy_hits = copy_number_file.get_single_copy_hits(db_genome_id)

        # gather information for all marker genes
        marker_paths = {"PFAM": os.path.join(self.pfam_hmm_dir, 'individual_hmms'),
                        "TIGRFAM": os.path.join(os.path.dirname(self.tigrfam_hmm_dir), 'individual_hmms')}

        marker_dict_original = {}
        if marker_set_id == "bac120":
            for db_marker in sorted(self.bac120_markers):
                marker_dict_original.update({marker.replace(".HMM", "").replace(".hmm", ""):
                                             os.path.join(
                    marker_paths[db_marker], marker)
                    for marker in self.bac120_markers[db_marker]})
        elif marker_set_id == "ar122":
            for db_marker in sorted(self.ar122_markers):
                marker_dict_original.update({marker.replace(".HMM", "").replace(".hmm", ""):
                                             os.path.join(
                    marker_paths[db_marker], marker)
                    for marker in self.ar122_markers[db_marker]})
        elif marker_set_id == "rps23":
            for db_marker in sorted(self.rps23_markers):
                marker_dict_original.update({marker.replace(".HMM", "").replace(".hmm", ""):
                                             os.path.join(
                    marker_paths[db_marker], marker)
                    for marker in self.rps23_markers[db_marker]})

        # Iterate over each of the expected markers and store the gene sequence.
        gene_dict = dict()
        result_align = dict()
        for marker_id, marker_path in marker_dict_original.items():
            hit = single_copy_hits.get(marker_id)
            if hit:
                # print(marker_id)
                gene_dict[marker_id] = {"marker_path": marker_path,
                                        "gene": hit['hit'].gene_id,
                                        "gene_seq": hit['seq'],
                                        "bitscore": hit['hit'].bit_score}
            else:
                hmm_len = self._get_hmm_size(marker_path)
                result_align[marker_id] = '-' * hmm_len

        # Align the markers.
        result_align.update(self._run_align(gene_dict, db_genome_id))

        # we concatenate the aligned markers together and associate them with
        # the genome.
        return ''.join([x[1] for x in sorted(result_align.items())])

    def _run_align(self, marker_dict, genome):
        """
        Run hmmalign for a set of genes for a specific genome. This is run in a temp folder.
        :param marker_dict: list of markers that need to be aligned
        :param genome: specific genome id
        Returns
        --------------
        List of tuple to be inserted in aligned_markers table
        """
        result_genomes_dict = {}
        hmmalign_dir = tempfile.mkdtemp()
        input_count = 0
        for markerid, marker_info in marker_dict.items():
            hmmalign_gene_input = os.path.join(
                hmmalign_dir, "input_gene{0}.fa".format(input_count))
            input_count += 1
            with open(hmmalign_gene_input, 'w') as out_fh:
                out_fh.write(">{0}\n".format(marker_info.get("gene")))
                out_fh.write("{0}".format(marker_info.get("gene_seq")))
            proc = subprocess.Popen(["hmmalign", "--outformat", "Pfam", marker_info.get(
                "marker_path"), hmmalign_gene_input], stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, encoding='utf-8')
            stdout, stderr = proc.communicate()

            for line in stderr.splitlines():
                print(line)

            result = self._get_aligned_marker(
                marker_info.get("gene"), stdout)
            if len(result) < 1:
                return "TODO"

            result_genomes_dict[markerid] = result

            input_count += 1
        shutil.rmtree(hmmalign_dir)
        return result_genomes_dict

    def _get_hmm_size(self, path):
        size = 0
        with open(path) as fp:
            for line in fp:
                if line.startswith("LENG  "):
                    size = line.split("  ")[1]
                    break
        return int(size)

    def _get_aligned_marker(self, hit_name, result_file):
        """
        Parse the output of Hmmalign
        :param hit_name: gene name
        :param result_file: output file from Hmmalign
        """
        hit_seq = None
        mask_seq = None

        for line in result_file.splitlines():
            splitline = line.split(" ", 1)
            if splitline[0] == hit_name.split(" ", 1)[0]:
                rsplitline = line.rsplit(" ", 1)
                hit_seq = rsplitline[-1]
                continue
            if line[0:len("#=GC RF")] == "#=GC RF":
                rsplitline = line.rsplit(" ", 1)
                mask_seq = rsplitline[-1]

        if mask_seq is None:
            raise Exception("Unable to get mask from hmm align result file")

        if hit_seq is None:
            return None

        aligned_marker = ""
        for pos in range(0, len(mask_seq)):
            if mask_seq[pos] != 'x':
                continue
            aligned_marker += hit_seq[pos]
        return aligned_marker



import hashlib
import math
import os
import random
import re
import time
from itertools import islice

import gtdbtk.config.config as Config
from gtdbtk.config.output import CHECKSUM_SUFFIX


##################################################
############MISC UTILITIES########################
##################################################


def get_reference_ids():
    results = []
    with open(Config.TAXONOMY_FILE) as tf:
        for line in tf:
            raw_id = line.split('\t')[0]
            results.append(raw_id)
            if raw_id[0:4] in ['GCF_', 'GCA_']:
                results.append(add_ncbi_prefix(raw_id))
            elif raw_id[0:3] in ['RS_', 'GB_']:
                results.append(raw_id[3:])

    return results


def add_ncbi_prefix(refname):
    if refname.startswith("GCF_"):
        return "RS_" + refname
    elif refname.startswith("GCA_"):
        return "GB_" + refname
    else:
        return refname


def splitchunks(d, n):
    chunksize = int(math.ceil(len(d) / float(n)))
    it = iter(d)
    for _ in range(0, len(d), chunksize):
        yield {k: d[k] for k in islice(it, chunksize)}


def splitchunks_list(l, n):
    """Yield successive n-sized chunks from l."""
    chunksize = int(math.ceil(len(l) / float(n)))
    for i in range(0, len(l), chunksize):
        yield l[i:i + chunksize]


def generateTempTableName():
    rng = random.SystemRandom()
    suffix = ''
    for _ in range(0, 10):
        suffix += rng.choice(
            'abcefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')
    return "TEMP" + suffix + str(int(time.time()))


def merge_two_dicts(x, y):
    """Given two dicts, merge them into a new dict as a shallow copy."""
    z = x.copy()
    z.update(y)
    return z


def sha1_dir(path, progress):
    """Recursively add files found within the path and output a SHA1 hash.

    Parameters
    ----------
    path : str
        The path to traverse.
    progress : bool
        True if progress should be displayed to stdout, False otherwise.

    Returns
    -------
    str
        The SHA1 hash representing this directory.
    """
    if progress:
        print('\r[{}]'.format(path), end='', flush=True)

    # Generate a queue of files to process
    queue = list()
    for root, dirs, files in os.walk(path):
        for file in files:
            path_file = os.path.join(root, file)
            queue.append(path_file)
    queue = sorted(queue)

    # Setup the hasher
    block_size = 65536
    hasher = hashlib.sha1()

    # Process the queue and obtain a single hash
    for idx, cur_path in enumerate(queue):

        if progress:
            print('\r[{}] - {}/{} files ({}%)'.format(path,
                                                      idx,
                                                      len(queue),
                                                      round(100 * (idx / len(queue)), 2)),
                  end='', flush=True)

        # Add the hash of the file
        with open(cur_path, 'rb') as fh:
            buf = fh.read(block_size)
            while len(buf) > 0:
                hasher.update(buf)
                buf = fh.read(block_size)

    if progress:
        print('\r' + ' ' * 100, end='', flush=True)
        print('\r', end='', flush=True)

    return hasher.hexdigest()


def sha256(input_file):
    """Determine SHA256 hash for file.

    Parameters
    ----------
    input_file : str
        Name of file.
    Returns
    -------
    str
        SHA256 hash.
    """
    block_size = 65536
    hasher = hashlib.sha1()
    with open(input_file, 'rb') as afile:
        buf = afile.read(block_size)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(block_size)
    return hasher.hexdigest()


def file_has_checksum(file_path, checksum_suffix=CHECKSUM_SUFFIX):
    """Check that the file contents match the checksum.

    Parameters
    ----------
    file_path : str
        Name of the file to check.
    checksum_suffix : str
        Suffix used to denote the file checksum.

    Returns
    -------
    bool
        True if the file has a checksum and it matches the original contents,
        False otherwise.

    """
    check_path = file_path + checksum_suffix
    if os.path.isfile(file_path) and os.path.isfile(check_path):
        with open(check_path, 'r') as check_f:
            return sha256(file_path) == check_f.read()
    return False


def symlink_f(src, dst, force=True):
    """Create a symbolic link pointing to src named dst.

    Parameters
    ----------
    src : str
        The source file.
    dst : str
        The destination file.
    force : bool
        Overwrite any file found with the same name as dst.

    """
    if force and os.path.isfile(dst):
        os.remove(dst)
    os.symlink(src, dst)


def get_memory_gb():
    try:
        with open('/proc/meminfo', 'r') as mem:
            hits = re.findall(r'([\w()]+):\s+(\d+) kB', mem.read())
        return {k: round(int(v) / 1e6, 2) for k, v in hits}
    except Exception:
        return None


def get_proc_memory_gb(pid):
    virt, res = None, None
    try:
        with open(f'/proc/{pid}/status', 'r') as fh:
            contents = fh.read()
        virt = int(re.search(r'VmSize:[^\d]+(\d+)', contents).group(1)) / 1e6
        res = int(re.search(r'VmRSS:[^\d]+(\d+)', contents).group(1)) / 1e6
    finally:
        return virt, res



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
import os
import shutil
import subprocess
import sys

import gtdbtk.config.config as Config
from gtdbtk.ani_rep import ANIRep
from gtdbtk.biolib_lite.common import (check_dir_exists,
                                       check_file_exists,
                                       make_sure_path_exists,
                                       remove_extension)
from gtdbtk.biolib_lite.execute import check_dependencies
from gtdbtk.biolib_lite.logger import colour
from gtdbtk.biolib_lite.taxonomy import Taxonomy
from gtdbtk.classify import Classify
from gtdbtk.config.output import *
from gtdbtk.decorate import Decorate
from gtdbtk.exceptions import *
from gtdbtk.external.fasttree import FastTree
from gtdbtk.infer_ranks import InferRanks
from gtdbtk.markers import Markers
from gtdbtk.misc import Misc
from gtdbtk.reroot_tree import RerootTree
from gtdbtk.tools import symlink_f, get_reference_ids


class OptionsParser(object):

    def __init__(self, version):
        """Initialization.

        Parameters
        ----------
        version : str
            The current version number (e.g. 0.2.2).
        """
        self.logger = logging.getLogger('timestamp')
        self.warnings = logging.getLogger('warnings')
        self.version = version
        self._check_package_compatibility()

        self.genomes_to_process = None

    def _check_package_compatibility(self):
        """Check that GTDB-Tk is using the most up-to-date reference package."""
        pkg_ver = float(Config.VERSION_DATA.replace('r', ''))
        min_ver = float(Config.MIN_REF_DATA_VERSION.replace('r', ''))
        self.logger.info('Using GTDB-Tk reference data version {}: {}'
                         .format(Config.VERSION_DATA, Config.GENERIC_PATH))
        if pkg_ver < min_ver:
            self.logger.warning(colour('You are not using the reference data '
                                       'intended for this release: {}'
                                       .format(Config.MIN_REF_DATA_VERSION),
                                       ['bright'], fg='yellow'))

    def _verify_genome_id(self, genome_id):
        """Ensure genome ID will be valid in Newick tree.

        Parameters
        ----------
        genome_id : str
            The string representing the genome identifier.

        Returns
        -------
        bool
            True if the genome identifier is legal.

        Raises
        ------
        GenomeNameInvalid
            If the genome identifier contains illegal characters.
        """

        invalid_chars = set('()[],;=')
        if any((c in invalid_chars) for c in genome_id):
            self.logger.error('Invalid genome ID: %s' % genome_id)
            self.logger.error(
                'The following characters are invalid: %s' % ' '.join(invalid_chars))
            raise GenomeNameInvalid('Invalid genome ID: {}'.format(genome_id))
        return True

    def _genomes_to_process(self, genome_dir, batchfile, extension):
        """Get genomes to process.

        Parameters
        ----------
        genome_dir : str
            Directory containing genomes.
        batchfile : str
            File describing genomes.
        extension : str
            Extension of files to process.

        Returns
        -------
        genomic_files : d[genome_id] -> FASTA file
            Map of genomes to their genomic FASTA files.
        """

        genomic_files, tln_tables = dict(), dict()
        if genome_dir:
            for f in os.listdir(genome_dir):
                if f.endswith(extension):
                    genome_id = remove_extension(f, extension)
                    genomic_files[genome_id] = os.path.join(genome_dir, f)

        elif batchfile:
            with open(batchfile, "r") as fh:
                for line_no, line in enumerate(fh):
                    line_split = line.strip().split("\t")
                    if line_split[0] == '':
                        continue  # blank line

                    if len(line_split) not in {2, 3}:
                        raise GTDBTkExit('Batch file must contain either 2 '
                                         'columns (detect translation table), '
                                         'or 3 (specify translation table).')

                    if len(line_split) == 2:
                        genome_file, genome_id = line_split
                    elif len(line_split) == 3:
                        genome_file, genome_id, tln_table = line_split
                        if tln_table not in {'4', '11'}:
                            raise GTDBTkExit('Specified translation table must '
                                             'be either 4, or 11.')
                        tln_tables[genome_id] = int(tln_table)

                    self._verify_genome_id(genome_id)

                    if genome_file is None or genome_file == '':
                        raise GTDBTkExit(
                            'Missing genome file on line %d.' % (line_no + 1))
                    elif genome_id is None or genome_id == '':
                        raise GTDBTkExit(
                            'Missing genome ID on line %d.' % (line_no + 1))
                    elif genome_id in genomic_files:
                        raise GTDBTkExit(
                            'Genome ID %s appears multiple times.' % genome_id)
                    if genome_file in genomic_files.values():
                        self.logger.warning(
                            'Genome file appears multiple times: %s' % genome_file)

                    genomic_files[genome_id] = genome_file

        # Check that the prefix is valid and the path exists
        invalid_paths = list()
        for genome_key, genome_path in genomic_files.items():

            if not os.path.isfile(genome_path):
                invalid_paths.append((genome_key, genome_path))

        # Report on any invalid paths
        if len(invalid_paths) > 0:
            self.warnings.info(f'Reading from batchfile: {batchfile}')
            self.warnings.error(f'The following {len(invalid_paths)} genomes '
                                f'have invalid paths specified in the batchfile:')
            for g_path, g_gid in invalid_paths:
                self.warnings.info(f'{g_gid}\t{g_path}')
            raise GTDBTkExit(f'There are {len(invalid_paths)} paths in the '
                             f'batchfile which do not exist, see gtdb.warnings.log')

        if len(genomic_files) == 0:
            if genome_dir:
                self.logger.error('No genomes found in directory: %s. Check '
                                  'the --extension flag used to identify '
                                  'genomes.' % genome_dir)
            else:
                self.logger.error('No genomes found in batch file: %s. Please '
                                  'check the format of this file.' % batchfile)
            raise GTDBTkExit

        invalid_genomes = set(genomic_files.keys()) & set(get_reference_ids())
        if len(invalid_genomes) > 0:
            self.warnings.info(f'The following {len(invalid_genomes)} have the '
                               f'same ID as GTDB-Tk reference genomes:')
            for invalid_genome in sorted(invalid_genomes):
                self.warnings.info(invalid_genome)
            raise GTDBTkExit(f'You have {len(invalid_genomes)} genomes with the '
                             f'same id as GTDB-Tk reference genomes, please '
                             f'rename them. See gtdb.warnings.log.')

        return genomic_files, tln_tables

    def _marker_set_id(self, bac120_ms, ar122_ms, rps23_ms):
        """Get unique identifier for marker set.

        Parameters
        ----------
        bac120_ms : bool
        ar122_ms : bool
        rps23_ms : bool

        Returns
        -------
        str
            The unique identifier for the marker set.

        Raises
        ------
        GenomeMarkerSetUnknown
            If the marker set is unknown.
        """

        if bac120_ms:
            marker_set_id = "bac120"
        elif ar122_ms:
            marker_set_id = "ar122"
        elif rps23_ms:
            marker_set_id = "rps23"
        else:
            self.logger.error('No marker set specified.')
            raise GenomeMarkerSetUnknown('No marker set specified.')
        return marker_set_id

    def _read_taxonomy_files(self, options):
        """Read and merge taxonomy files."""

        self.logger.info('Reading GTDB taxonomy for representative genomes.')
        taxonomy = Taxonomy().read(Config.TAXONOMY_FILE)

        if options.gtdbtk_classification_file:
            # add and overwrite taxonomy for genomes specified in the
            # GTDB-Tk classification file
            check_file_exists(options.gtdbtk_classification_file)

            self.logger.info('Reading GTDB-Tk classification file.')
            gtdbtk_taxonomy = Taxonomy().read(options.gtdbtk_classification_file)
            del gtdbtk_taxonomy['user_genome']
            num_reassigned = 0
            for gid, taxa in gtdbtk_taxonomy.items():
                if gid in taxonomy:
                    num_reassigned += 1
                taxonomy[gid] = taxa

            self.logger.info(f'Read GTDB-Tk classifications for {len(gtdbtk_taxonomy):,} genomes.')
            self.logger.info(f'Reassigned taxonomy for {num_reassigned:,} GTDB representative genomes.')

        if options.custom_taxonomy_file:
            # add and overwrite taxonomy for genomes specified in the
            # custom taxonomy file
            check_file_exists(options.custom_taxonomy_file)

            self.logger.info('Reading custom taxonomy file.')
            custom_taxonomy = Taxonomy().read(options.custom_taxonomy_file)
            num_reassigned = 0
            for gid, taxa in custom_taxonomy.items():
                if gid in taxonomy:
                    num_reassigned += 1
                taxonomy[gid] = taxa

            self.logger.info(f'Read custom taxonomy for {len(custom_taxonomy):,} genomes.')
            self.logger.info(f'Reassigned taxonomy for {num_reassigned:,} GTDB representative genomes.')

        if options.gtdbtk_classification_file and options.custom_taxonomy_file:
            dup_genomes = set(gtdbtk_taxonomy).intersection(custom_taxonomy)
            if len(dup_genomes) > 0:
                self.logger.error('GTDB-Tk classification and custom taxonomy '
                                  'files must not specify taxonomies for the '
                                  'same genomes.')
                self.logger.error('These files have {:,} genomes in common.'.format(len(dup_genomes)))
                self.logger.error('Example duplicate genome: {}'.format(dup_genomes.pop()))
                raise GTDBTkExit('Duplicated taxonomy information.')

        self.logger.info(f'Read taxonomy for {len(taxonomy):,} genomes.')

        return taxonomy

    def identify(self, options):
        """Identify marker genes in genomes.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """

        if options.genome_dir:
            check_dir_exists(options.genome_dir)

        if options.batchfile:
            check_file_exists(options.batchfile)

        make_sure_path_exists(options.out_dir)

        genomes, tln_tables = self._genomes_to_process(
            options.genome_dir, options.batchfile, options.extension)
        self.genomes_to_process = genomes

        markers = Markers(options.cpus)
        markers.identify(genomes,
                         tln_tables,
                         options.out_dir,
                         options.prefix,
                         options.force)

        self.logger.info('Done.')

    def align(self, options):
        """Create MSA from marker genes.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """

        check_dir_exists(options.identify_dir)
        make_sure_path_exists(options.out_dir)

        if not hasattr(options, 'outgroup_taxon'):
            options.outgroup_taxon = None

        markers = Markers(options.cpus, options.debug)
        markers.align(options.identify_dir,
                      options.skip_gtdb_refs,
                      options.taxa_filter,
                      options.min_perc_aa,
                      options.custom_msa_filters,
                      options.skip_trimming,
                      options.rnd_seed,
                      options.cols_per_gene,
                      options.min_consensus,
                      options.max_consensus,
                      options.min_perc_taxa,
                      options.out_dir,
                      options.prefix,
                      options.outgroup_taxon,
                      self.genomes_to_process)

        self.logger.info('Done.')

    def infer(self, options):
        """Infer a tree from a user specified MSA.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """

        check_file_exists(options.msa_file)
        make_sure_path_exists(options.out_dir)

        if options.cpus > 1:
            check_dependencies(['FastTreeMP'])
        else:
            check_dependencies(['FastTree'])

        if hasattr(options, 'suffix'):
            output_tree = os.path.join(options.out_dir,
                                       PATH_MARKER_UNROOTED_TREE.format(prefix=options.prefix,
                                                                        marker=options.suffix))
            tree_log = os.path.join(options.out_dir,
                                    PATH_MARKER_TREE_LOG.format(prefix=options.prefix,
                                                                marker=options.suffix))
            fasttree_log = os.path.join(options.out_dir,
                                        PATH_MARKER_FASTTREE_LOG.format(prefix=options.prefix,
                                                                        marker=options.suffix))
        else:
            output_tree = os.path.join(options.out_dir,
                                       PATH_UNROOTED_TREE.format(prefix=options.prefix))
            tree_log = os.path.join(options.out_dir,
                                    PATH_TREE_LOG.format(prefix=options.prefix))
            fasttree_log = os.path.join(options.out_dir,
                                        PATH_FASTTREE_LOG.format(prefix=options.prefix))

        fasttree = FastTree()
        fasttree.run(output_tree, tree_log, fasttree_log, options.prot_model,
                     options.no_support, options.gamma, options.msa_file,
                     options.cpus)
        self.logger.info(f'FastTree version: {fasttree.version}')

        if hasattr(options, 'subparser_name') and options.subparser_name == 'infer':
            symlink_f(output_tree[len(options.out_dir) + 1:],
                      os.path.join(options.out_dir,
                                   os.path.basename(output_tree)))

        self.logger.info('Done.')

    def run_test(self, options):
        """Run test of classify workflow.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.

        Returns
        -------
        bool
            True if the test succeeds.

        Raises
        ------
        GTDBTkTestFailure
            If the test fails.
        """

        make_sure_path_exists(options.out_dir)

        output_dir = os.path.join(options.out_dir, 'output')
        genome_test_dir = os.path.join(options.out_dir, 'genomes')
        if os.path.exists(genome_test_dir):
            self.logger.error(
                'Test directory {} already exists'.format(genome_test_dir))
            self.logger.error('Test must be run in a new directory.')
            sys.exit(1)

        current_path = os.path.dirname(os.path.realpath(__file__))
        input_dir = os.path.join(current_path, 'tests', 'data', 'genomes')

        shutil.copytree(input_dir, genome_test_dir)

        args = ['gtdbtk', 'classify_wf', '--genome_dir', genome_test_dir,
                '--out_dir', output_dir, '--cpus', str(options.cpus)]
        self.logger.info('Command: {}'.format(' '.join(args)))

        path_stdout = os.path.join(options.out_dir, 'test_execution.log')
        with open(path_stdout, 'w') as fh_stdout:
            proc = subprocess.Popen(args, stdout=fh_stdout,
                                    stderr=subprocess.PIPE)
            proc.communicate()

        summary_file = os.path.join(
            output_dir, PATH_AR122_SUMMARY_OUT.format(prefix='gtdbtk'))

        if proc.returncode != 0:
            self.logger.error('The test returned a non-zero exit code.')
            self.logger.error('A detailed summary of the execution log can be '
                              'found here: {}'.format(path_stdout))
            self.logger.error('The test has failed.')
            sys.exit(1)
        if not os.path.exists(summary_file):
            self.logger.error("{} is missing.".format(summary_file))
            self.logger.error('A detailed summary of the execution log can be '
                              'found here: {}'.format(path_stdout))
            self.logger.error('The test has failed.')
            sys.exit(1)

        self.logger.info('Test has successfully finished.')
        return True

    def classify(self, options):
        """Determine taxonomic classification of genomes.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """

        # See ticket #255... perhaps an upstream version/OS issue?
        if not hasattr(options, 'pplacer_cpus'):
            options.pplacer_cpus = None

        check_dir_exists(options.align_dir)
        make_sure_path_exists(options.out_dir)
        if options.scratch_dir:
            make_sure_path_exists(options.scratch_dir)

        genomes, _ = self._genomes_to_process(
            options.genome_dir, options.batchfile, options.extension)

        classify = Classify(options.cpus, options.pplacer_cpus)
        classify.run(genomes,
                     options.align_dir,
                     options.out_dir,
                     options.prefix,
                     options.scratch_dir,
                     options.recalculate_red,
                     options.debug,
                     options.split_tree)

        self.logger.info('Done.')

    def trim_msa(self, options):
        """ Trim an untrimmed archaea or bacterial MSA file.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """
        if options.reference_mask in ['bac', 'arc']:
            mask_type = "reference"
            mask_id = options.reference_mask
        else:
            mask_type = "file"
            mask_id = options.mask_file
        misc = Misc()
        misc.trim_msa(options.untrimmed_msa, mask_type,
                      mask_id, options.output)
        self.logger.info('Done.')

    def export_msa(self, options):
        """Export the untrimmed archaeal or bacterial MSA file.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """
        misc = Misc()
        misc.export_msa(options.domain, options.output)

        self.logger.info('Done.')

    def root(self, options):
        """Root tree using outgroup.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """

        check_file_exists(options.input_tree)

        taxonomy = self._read_taxonomy_files(options)

        self.logger.info(f'Identifying genomes from the specified outgroup: {options.outgroup_taxon}')
        outgroup = set()
        for genome_id, taxa in taxonomy.items():
            if options.outgroup_taxon in taxa:
                outgroup.add(genome_id)

        reroot = RerootTree()
        reroot.root_with_outgroup(options.input_tree,
                                  options.output_tree,
                                  outgroup)

        self.logger.info('Done.')

    def decorate(self, options):
        """Decorate tree with GTDB taxonomy.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """

        check_file_exists(options.input_tree)

        taxonomy = self._read_taxonomy_files(options)

        d = Decorate()
        d.run(options.input_tree,
              taxonomy,
              options.output_tree)

        self.logger.info('Done.')

        # symlink to the decorated tree file, if not run independently
        if hasattr(options, 'suffix'):
            if options.suffix == 'bac120':
                symlink_f(PATH_BAC120_DECORATED_TREE.format(prefix=options.prefix),
                          os.path.join(options.out_dir,
                                       os.path.basename(PATH_BAC120_DECORATED_TREE.format(prefix=options.prefix))))
                symlink_f(PATH_BAC120_DECORATED_TREE.format(prefix=options.prefix) + '-table',
                          os.path.join(options.out_dir,
                                       os.path.basename(PATH_BAC120_DECORATED_TREE.format(prefix=options.prefix)  + '-table')))
            elif options.suffix == 'ar122':
                symlink_f(PATH_AR122_DECORATED_TREE.format(prefix=options.prefix),
                          os.path.join(options.out_dir,
                                       os.path.basename(PATH_AR122_DECORATED_TREE.format(prefix=options.prefix))))
                symlink_f(PATH_AR122_DECORATED_TREE.format(prefix=options.prefix) + '-table',
                          os.path.join(options.out_dir,
                                       os.path.basename(PATH_AR122_DECORATED_TREE.format(prefix=options.prefix)  + '-table')))
            else:
                raise GenomeMarkerSetUnknown(
                    'There was an error determining the marker set.')

    def check_install(self):
        """ Verify all GTDB-Tk data files are present.

        Raises
        ------
        ReferenceFileMalformed
            If one or more reference files are malformed.
        """
        self.logger.info("Running install verification")
        misc = Misc()
        misc.check_install()
        self.logger.info('Done.')

    def infer_ranks(self, options):
        """Establish taxonomic ranks of internal nodes using RED."""

        check_file_exists(options.input_tree)

        p = InferRanks()
        p.run(options.input_tree,
              options.ingroup_taxon,
              options.output_tree)

        self.logger.info('Done.')

    def ani_rep(self, options):
        """Calculates ANI to GTDB representative genomes.

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """
        make_sure_path_exists(options.out_dir)

        genomes, _ = self._genomes_to_process(
            options.genome_dir, options.batchfile, options.extension)

        ani_rep = ANIRep(options.cpus)
        ani_rep.run(genomes, options.no_mash, options.mash_d, options.out_dir, options.prefix,
                    options.mash_k, options.mash_v, options.mash_s, options.min_af)

        self.logger.info('Done.')

    def parse_options(self, options):
        """Parse user options and call the correct pipeline(s)

        Parameters
        ----------
        options : argparse.Namespace
            The CLI arguments input by the user.
        """

        # Stop processing if python 2 is being used.
        if sys.version_info.major < 3:
            raise GTDBTkExit('Python 2 is no longer supported.')

        # Correct user paths
        if hasattr(options, 'out_dir'):
            options.out_dir = os.path.expanduser(options.out_dir)

        # Assert that the number of CPUs is a positive integer.
        if hasattr(options, 'cpus') and options.cpus < 1:
            self.logger.warning(
                'You cannot use less than 1 CPU, defaulting to 1.')
            options.cpus = 1

        if options.subparser_name == 'de_novo_wf':
            check_dependencies(['prodigal', 'hmmalign'])
            check_dependencies(
                ['FastTree' + ('MP' if options.cpus > 1 else '')])

            self.identify(options)

            options.identify_dir = options.out_dir
            options.skip_trimming = False
            self.align(options)

            if options.bacteria:
                options.suffix = "bac120"
            else:
                options.suffix = "ar122"

            if options.skip_gtdb_refs:
                if options.suffix == 'bac120':
                    options.msa_file = os.path.join(
                        options.out_dir, PATH_BAC120_USER_MSA.format(prefix=options.prefix))
                elif options.suffix == 'ar122':
                    options.msa_file = os.path.join(
                        options.out_dir, PATH_AR122_USER_MSA.format(prefix=options.prefix))
                else:
                    self.logger.error(
                        'There was an error determining the marker set.')
                    raise GenomeMarkerSetUnknown(
                        'Unknown marker set: {}'.format(options.suffix))
            else:
                if options.suffix == 'bac120':
                    options.msa_file = os.path.join(
                        options.out_dir, PATH_BAC120_MSA.format(prefix=options.prefix))
                elif options.suffix == 'ar122':
                    options.msa_file = os.path.join(
                        options.out_dir, PATH_AR122_MSA.format(prefix=options.prefix))
                else:
                    self.logger.error(
                        'There was an error determining the marker set.')
                    raise GenomeMarkerSetUnknown(
                        'Unknown marker set: {}'.format(options.suffix))

            self.infer(options)

            if options.suffix == 'bac120':
                options.input_tree = os.path.join(options.out_dir,
                                                  PATH_BAC120_UNROOTED_TREE.format(prefix=options.prefix))
                options.output_tree = os.path.join(options.out_dir,
                                                   PATH_BAC120_ROOTED_TREE.format(prefix=options.prefix))
            elif options.suffix == 'ar122':
                options.input_tree = os.path.join(options.out_dir,
                                                  PATH_AR122_UNROOTED_TREE.format(prefix=options.prefix))
                options.output_tree = os.path.join(options.out_dir,
                                                   PATH_AR122_ROOTED_TREE.format(prefix=options.prefix))

            self.root(options)

            if options.suffix == 'bac120':
                options.input_tree = os.path.join(options.out_dir,
                                                  PATH_BAC120_ROOTED_TREE.format(prefix=options.prefix))
                options.output_tree = os.path.join(options.out_dir,
                                                   PATH_BAC120_DECORATED_TREE.format(prefix=options.prefix))
            elif options.suffix == 'ar122':
                options.input_tree = os.path.join(options.out_dir,
                                                  PATH_AR122_ROOTED_TREE.format(prefix=options.prefix))
                options.output_tree = os.path.join(options.out_dir,
                                                   PATH_AR122_DECORATED_TREE.format(prefix=options.prefix))

            self.decorate(options)

        elif options.subparser_name == 'classify_wf':

            # TODO: Remove this block once the split_tree function is implemented.
            if hasattr(options, 'split_tree'):
                self.logger.warning('The split tree option is not yet '
                                    ' supported, overriding value to False.')
            options.split_tree = False

            check_dependencies(
                ['prodigal', 'hmmalign', 'pplacer', 'guppy', 'fastANI'])
            self.identify(options)

            options.identify_dir = options.out_dir
            options.align_dir = options.out_dir
            options.taxa_filter = None
            options.custom_msa_filters = False
            # Added here due to the other mutex argument being include above.
            options.skip_trimming = False
            options.min_consensus = None
            options.min_perc_taxa = None
            options.skip_gtdb_refs = False
            options.cols_per_gene = None
            options.max_consensus = None
            options.rnd_seed = None
            options.skip_trimming = False

            self.align(options)

            self.classify(options)
        elif options.subparser_name == 'identify':
            self.identify(options)
        elif options.subparser_name == 'align':
            self.align(options)
        elif options.subparser_name == 'infer':
            self.infer(options)
        elif options.subparser_name == 'classify':

            # TODO: Remove this block once the split_tree function is implemented.
            if hasattr(options, 'split_tree'):
                self.logger.warning('The split tree option is not yet '
                                    ' supported, overriding value to False.')
            options.split_tree = False

            if options.recalculate_red and options.split_tree:
                raise GTDBTkExit('--split_tree and --recalculate_red are mutually exclusive.')
            self.classify(options)
        elif options.subparser_name == 'root':
            self.root(options)
        elif options.subparser_name == 'decorate':
            self.decorate(options)
        elif options.subparser_name == 'infer_ranks':
            self.infer_ranks(options)
        elif options.subparser_name == 'ani_rep':
            self.ani_rep(options)
        elif options.subparser_name == 'trim_msa':
            self.trim_msa(options)
        elif options.subparser_name == 'export_msa':
            self.export_msa(options)
        elif options.subparser_name == 'test':
            check_dependencies(['prodigal', 'hmmalign', 'pplacer', 'guppy',
                                'fastANI'])
            self.run_test(options)
        elif options.subparser_name == 'check_install':
            self.check_install()
        else:
            self.logger.error('Unknown GTDB-Tk command: "' +
                              options.subparser_name + '"\n')
            sys.exit(1)

        return 0



#!/usr/bin/env python

###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################


__prog_name__ = 'trim_msa.py'
__prog_desc__ = 'Randomly select a subset of columns from the MSA of each marker.'

__author__ = 'Pierre Chaumeil and Donovan Parks'
__copyright__ = 'Copyright 2017'
__credits__ = ['Pierre Chaumeil', 'Donovan Parks']
__license__ = 'GPL3'
__version__ = '0.0.4'
__maintainer__ = 'Pierre Chaumeil'
__email__ = 'p.chaumeil@uq.edu.au'
__status__ = 'Development'

import argparse
import logging
import os
import random
import sys
from collections import defaultdict, Counter

from numpy import (mean as np_mean,
                   std as np_std)

from gtdbtk.biolib_lite.logger import logger_setup
from gtdbtk.biolib_lite.seq_io import read_fasta
from gtdbtk.exceptions import MSAMarkerLengthMismatch


class TrimMSA(object):
    """Randomly select a subset of columns from the MSA of each marker."""

    def __init__(self, cols_per_gene,
                 min_perc_aa,
                 min_consensus,
                 max_consensus,
                 min_perc_taxa,
                 rnd_seed,
                 out_dir):
        """Initialization."""

        self.output_dir = out_dir
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        self.subset = cols_per_gene  # default: 42 * ~120 genes ~= 5,000 columns

        # only consider columns with less than this percentage of gaps
        self.max_gaps = 1.0 - min_perc_taxa

        # only consider columns where the most common amino acid is
        # between these two percent boundaries
        self.min_identical_aa = min_consensus
        self.max_identical_aa = max_consensus

        # remove genomes without sufficient number of amino acids in MSA
        self.min_perc_aa = min_perc_aa

        random.seed(rnd_seed)

        self.logger = logging.getLogger('timestamp')

    def run(self, msa_file, marker_list):
        """Randomly select a subset of columns from the MSA of each marker."""

        # read multiple sequence alignment
        self.logger.info('Reading multiple sequence alignment.')
        msa = read_fasta(msa_file, False)
        self.logger.info('Read MSA for %d genomes.' % len(msa))

        filtered_seqs, pruned_seqs = self.trim(msa, marker_list)

        self.logger.info('Removed %d taxa have amino acids in <%.1f%% of columns in filtered MSA.' % (
            len(pruned_seqs),
            self.min_perc_aa))

        # write out trimmed sequences
        with open(os.path.join(self.output_dir, "filtered_msa.faa"), 'w') as filter_file:
            for gid, seq in filtered_seqs.items():
                fasta_outstr = ">%s\n%s\n" % (gid, seq)
                filter_file.write(fasta_outstr)

        self.logger.info('Done.')

    def trim(self, msa, marker_list):
        """ Randomly select a subset of columns from the MSA of each marker.
        Args:
            msa (dict): The multiple sequence alignment to trim.
            marker_list (list): A list of markers to use for trimming (id/name/length).

        Returns:
            (dict, dict): The trimmed MSA, and any sequences which were omitted.

        Raises:
            MSAMarkerLengthMismatch: If the MSA length does not equal the length of the marker genes.
        """
        # get marker info
        self.logger.info('Reading marker info.')
        markers = []
        total_msa_len = 0
        with open(marker_list, 'r') as f:
            f.readline()
            for line in f:
                list_info = line.split("\t")
                marker_id = list_info[0]
                marker_name = '%s: %s' % (list_info[1], list_info[2])
                marker_len = int(list_info[3])
                markers.append((marker_id, marker_name, marker_len))
                total_msa_len += marker_len

        if len(list(msa.values())[0]) == total_msa_len:
            self.logger.info(
                'Length of MSA and length of marker genes both equal %d columns' % total_msa_len)
        else:
            self.logger.error('Length of MSA (%d columns) does not equal length of marker genes (%d columns).' % (
                len(list(msa.values())[0]),
                total_msa_len))
            raise MSAMarkerLengthMismatch

        # randomly select columns meeting filtering criteria
        self.logger.info(
            'Randomly sampling %d columns passing filtering criteria from each marker gene.' % self.subset)
        mask, output_seqs = self.subsample_msa(msa, markers)

        # write mask to file
        with open(os.path.join(self.output_dir, "mask.txt"), 'w') as mask_file:
            mask_file.write(''.join([str(n) for n in mask]))

        # write subsampled MSA to file
        nbr_aa_seqs = open(os.path.join(
            self.output_dir, "genome_msa_stats.tsv"), 'w')
        nbr_aa_seqs.write(
            'Genome ID\tMSA length\tAmino acids\tAmino acids (%)\n')
        filtered_msa = {}
        pruned_seqs = {}
        for genome_id, aligned_seq in output_seqs.items():
            aa_len = sum([1 for c in aligned_seq if c.isalpha()])
            if aa_len != 0:
                aa_perc = float(aa_len) / len(aligned_seq)
            else:
                aa_perc = 0
            len_outstr = "%s\t%d\t%d\t%.2f\n" % (
                genome_id,
                len(aligned_seq),
                aa_len,
                aa_perc * 100.0)
            nbr_aa_seqs.write(len_outstr)

            if aa_perc >= self.min_perc_aa:
                filtered_msa[genome_id] = aligned_seq
            else:
                pruned_seqs[genome_id] = aligned_seq

        nbr_aa_seqs.close()

        return filtered_msa, pruned_seqs

    def identify_valid_columns(self, start, end, seqs):
        # type: (int, int, dict) -> set
        """Identify columns meeting gap and amino acid ubiquity criteria."""

        GAP_CHARS = {'-', '.', '_', '*'}
        STANDARD_AMINO_ACIDS = set('ACDEFGHIKLMNPQRSTVWY')

        gap_count = defaultdict(int)
        amino_acids = [list() for _ in range(end - start)]
        num_genomes = 0
        for seq_id, seq in seqs.items():
            num_genomes += 1
            gene_seq = seq[start:end].upper()
            for i, ch in enumerate(gene_seq):
                if ch in GAP_CHARS:
                    gap_count[i] += 1
                else:
                    amino_acids[i].append(ch)

        valid_cols = set()
        for i in range(0, end - start):
            if float(gap_count.get(i, 0)) / num_genomes <= self.max_gaps:
                c = Counter(amino_acids[i])
                if not c.most_common(1):
                    continue

                letter, count = c.most_common(1)[0]
                if letter not in STANDARD_AMINO_ACIDS:
                    self.logger.warning(
                        'Most common amino acid was not in standard alphabet: %s' % letter)

                aa_ratio = float(count) / (num_genomes - gap_count.get(i, 0))
                if self.min_identical_aa <= aa_ratio < self.max_identical_aa:
                    valid_cols.add(i)

        return valid_cols

    def subsample_msa(self, seqs, markers):
        # type: (dict, list) -> (list, dict)
        """Sample columns from each marker in multiple sequence alignment."""

        alignment_length = len(list(seqs.values())[0])
        sampled_cols = []
        start = 0
        lack_sufficient_cols = 0
        lack_cols_marker_ids = []
        avg_perc_cols = []
        for marker_id, marker_name, marker_len in markers:
            end = start + marker_len

            valid_cols = self.identify_valid_columns(start,
                                                     end,
                                                     seqs)
            assert (len(valid_cols) <= marker_len)  # sanity check

            self.logger.info('%s: S:%d, E:%d, LEN:%d, COLS:%d, PERC:%.1f' % (
                marker_name,
                start,
                end,
                marker_len,
                len(valid_cols),
                len(valid_cols) * 100.0 / marker_len))

            avg_perc_cols.append(len(valid_cols) * 100.0 / marker_len)

            if len(valid_cols) < self.subset:
                self.logger.warning(
                    'Marker has <%d columns after filtering.' % self.subset)
                lack_sufficient_cols += 1
                lack_cols_marker_ids.append(marker_id)

            offset_valid_cols = [i + start for i in valid_cols]
            sel_cols = random.sample(offset_valid_cols, min(
                self.subset, len(offset_valid_cols)))
            sampled_cols.extend(sel_cols)

            start = end

        mask = [1 if i in sampled_cols else 0 for i in range(alignment_length)]

        self.logger.info('Identified %d of %d marker genes with <%d columns for sampling:' % (
            lack_sufficient_cols,
            len(markers),
            self.subset))
        self.logger.info('%s' % ', '.join(lack_cols_marker_ids))
        self.logger.info('Marker genes had %.1f+/-%.1f%% of columns available for selection on average.' % (
            np_mean(avg_perc_cols),
            np_std(avg_perc_cols)))
        self.logger.info('Final MSA contains %d columns.' % len(sampled_cols))

        # trim columns
        output_seqs = {}
        for seq_id, seq in seqs.items():
            masked_seq = ''.join([seq[i]
                                  for i in range(0, len(mask)) if mask[i]])
            output_seqs[seq_id] = masked_seq

        return mask, output_seqs


if __name__ == '__main__':
    print(__prog_name__ + ' v' + __version__ + ': ' + __prog_desc__)
    print('  by ' + __author__ + ' (' + __email__ + ')' + '\n')

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--msa', help='unfiltered multiple sequence alignment')
    parser.add_argument(
        '--marker_list', help='file with metadata for each marker gene')
    parser.add_argument('--cols_per_gene', type=int, default=42,
                        help='maximum number of columns to retain per gene')
    parser.add_argument('--min_perc_aa', type=float, default=0.5,
                        help='filter genomes with an insufficient percentage of AA in the MSA')
    parser.add_argument('--min_consensus', type=float, default=0.25,
                        help='minimum percentage of the same amino acid required to retain column')
    parser.add_argument('--max_consensus', type=float, default=0.95,
                        help='maximum percentage of the same amino acid required to retain column')
    parser.add_argument('--min_perc_taxa', type=float, default=0.50,
                        help='minimum percentage of taxa required to retain column')
    parser.add_argument('--rnd_seed', type=int, default=None,
                        help='random seed to use for selecting columns')
    parser.add_argument('--out_dir', help='output directory')

    args = parser.parse_args()

    logger_setup(args.out_dir, "trim_msa.log", "trim_msa", __version__, False)

    try:
        p = TrimMSA(args.cols_per_gene,
                    args.min_perc_aa,
                    args.min_consensus,
                    args.max_consensus,
                    args.min_perc_taxa,
                    args.rnd_seed,
                    args.out_dir)
        p.run(args.msa, args.marker_list)
    except SystemExit:
        print("\nControlled exit resulting from an unrecoverable error or warning.")
    except:
        print("\nUnexpected error:", sys.exc_info()[0])
        raise



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import argparse
import logging
import sys
import traceback

from gtdbtk import __author__, __copyright__, __version__
from gtdbtk.biolib_lite.custom_help_formatter import CustomHelpFormatter
from gtdbtk.biolib_lite.exceptions import BioLibError
from gtdbtk.biolib_lite.logger import logger_setup
from gtdbtk.config.config import AF_THRESHOLD
from gtdbtk.exceptions import *
from gtdbtk.main import OptionsParser


def print_help():
    print('''\

              ...::: GTDB-Tk v%s :::...

  Workflows:
    classify_wf -> Classify genomes by placement in GTDB reference tree
                     (identify -> align -> classify)
    de_novo_wf  -> Infer de novo tree and decorate with GTDB taxonomy
                     (identify -> align -> infer -> root -> decorate)

  Methods:
    identify -> Identify marker genes in genome
    align    -> Create multiple sequence alignment
    classify -> Determine taxonomic classification of genomes
    infer    -> Infer tree from multiple sequence alignment
    root     -> Root tree using an outgroup
    decorate -> Decorate tree with GTDB taxonomy

  Tools:
    infer_ranks -> Establish taxonomic ranks of internal nodes using RED
    ani_rep     -> Calculates ANI to GTDB representative genomes
    trim_msa    -> Trim an untrimmed MSA file based on a mask
    export_msa  -> Export the untrimmed archaeal or bacterial MSA file

  Testing:
    test          -> Validate the classify_wf pipeline with 3 archaeal genomes 
    check_install -> Verify if all GTDB-Tk data files are present

  Use: gtdbtk <command> -h for command specific help
    ''' % __version__)


def main():
    parser = argparse.ArgumentParser(prog='gtdbtk', add_help=False, conflict_handler='resolve')
    parser.add_argument('-f', '--force', action="store_true", default=False,
                        help="overwrite existing files without prompting.")

    subparsers = parser.add_subparsers(help="--", dest='subparser_name')

    # de novo workflow
    denovo_wf_parser = subparsers.add_parser('de_novo_wf', conflict_handler='resolve',
                                             formatter_class=CustomHelpFormatter,
                                             help='Infer de novo tree and decorate with GTDB taxonomy.')

    mutual_genome_denovo_wf = denovo_wf_parser.add_argument_group('mutually exclusive required arguments')
    mutex_group = mutual_genome_denovo_wf.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument('--genome_dir',
                             help="directory containing genome files in FASTA format")
    mutex_group.add_argument('--batchfile',
                             help="file describing genomes - tab separated in 2 columns (FASTA file, genome ID)")

    mutual_ms_denovo_wf = denovo_wf_parser.add_argument_group('mutually exclusive required arguments')
    mutex_group = mutual_ms_denovo_wf.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument('--bacteria', action='store_true', help='process bacterial genomes')
    mutex_group.add_argument('--archaea', action='store_true', help='process archaeal genomes')

    required_denovo_wf = denovo_wf_parser.add_argument_group('required named arguments')
    required_denovo_wf.add_argument('--outgroup_taxon', required=True,
                                    help="taxon to use as outgroup (e.g., p__Patescibacteria or p__Altiarchaeota)")
    required_denovo_wf.add_argument('--out_dir', required=True,
                                    help="directory to output files")

    optional_denovo_wf = denovo_wf_parser.add_argument_group('optional arguments')
    optional_denovo_wf.add_argument('-x', '--extension', default='fna',
                                    help='extension of files to process, gz = gzipped')

    optional_denovo_wf.add_argument('--skip_gtdb_refs',
                                    action="store_true",
                                    help='do not include GTDB reference genomes in multiple sequence alignment')
    optional_denovo_wf.add_argument('--taxa_filter',
                                    help=('filter GTDB genomes to taxa (comma separated) within '
                                          + 'specific taxonomic groups (e.g., d__Bacteria '
                                          + 'or p__Proteobacteria, p__Actinobacteria)'))
    optional_denovo_wf.add_argument('--min_perc_aa', type=float, default=10,
                                    help='filter genomes with an insufficient percentage of AA in the MSA (inclusive bound)')
    optional_denovo_wf.add_argument('--custom_msa_filters', action="store_true",
                                    help=('perform custom filtering of MSA with cols_per_gene, min_consensus '
                                          + 'max_consensus, and min_perc_taxa parameters instead of using canonical mask'))
    optional_denovo_wf.add_argument('--cols_per_gene', type=int, default=42,
                                    help='maximum number of columns to retain per gene')
    optional_denovo_wf.add_argument('--min_consensus', type=float, default=25,
                                    help='minimum percentage of the same amino acid required to retain column (inclusive bound)')
    optional_denovo_wf.add_argument('--max_consensus', type=float, default=95,
                                    help='maximum percentage of the same amino acid required to retain column (exclusive bound)')
    optional_denovo_wf.add_argument('--min_perc_taxa', type=float, default=50,
                                    help='minimum percentage of taxa required to retain column (inclusive bound)')
    optional_denovo_wf.add_argument('--rnd_seed', type=int, default=None,
                                    help='random seed to use for selecting columns')

    optional_denovo_wf.add_argument('--prot_model', choices=['JTT', 'WAG', 'LG'],
                                    help='protein substitution model for tree inference', default='WAG')
    optional_denovo_wf.add_argument('--no_support', action="store_true",
                                    help="do not compute local support values using the Shimodaira-Hasegawa test")
    optional_denovo_wf.add_argument('--gamma', action="store_true",
                                    help="rescale branch lengths to optimize the Gamma20 likelihood")

    optional_denovo_wf.add_argument('--gtdbtk_classification_file',
                                    help="file with GTDB-Tk classifications produced by the `classify` command")
    optional_denovo_wf.add_argument('--custom_taxonomy_file',
                                    help="file indicating custom taxonomy string for at least the genomes belonging to the outgroup")

    optional_denovo_wf.add_argument('--prefix', default='gtdbtk',
                                    help='desired prefix for output files')
    optional_denovo_wf.add_argument('--cpus', default=1, type=int,
                                    help='number of CPUs to use')
    optional_denovo_wf.add_argument('--force', action='store_const', const=True, default=False,
                                    help='continue processing if an error occurs on a single genome')
    optional_denovo_wf.add_argument('--debug', action="store_true",
                                    help='create intermediate files for debugging purposes')
    optional_denovo_wf.add_argument('-h', '--help', action="help",
                                    help="show help message")

    # classify workflow
    classify_wf_parser = subparsers.add_parser('classify_wf', conflict_handler='resolve',
                                               formatter_class=CustomHelpFormatter,
                                               help='Classify genomes by placement in GTDB reference tree.')

    mutual_genome_classify_wf = classify_wf_parser.add_argument_group('mutually exclusive required arguments')
    mutex_group = mutual_genome_classify_wf.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument('--genome_dir',
                             help="directory containing genome files in FASTA format")
    mutex_group.add_argument('--batchfile',
                             help="file describing genomes - tab separated in 3 columns (FASTA file, genome ID, translation table [optional])")

    required_classify_wf = classify_wf_parser.add_argument_group('required named arguments')
    required_classify_wf.add_argument('--out_dir', required=True,
                                      help="directory to output files")

    optional_classify_wf = classify_wf_parser.add_argument_group('optional arguments')
    optional_classify_wf.add_argument('-x', '--extension', default='fna',
                                      help='extension of files to process, gz = gzipped')
    optional_classify_wf.add_argument('--min_perc_aa', type=float, default=10,
                                      help='filter genomes with an insufficient percentage of AA in the MSA')
    optional_classify_wf.add_argument('--prefix', required=False, default='gtdbtk',
                                      help='desired prefix for output files')
    optional_classify_wf.add_argument('--cpus', default=1, type=int,
                                      help='number of CPUs to use')
    optional_classify_wf.add_argument('--pplacer_cpus', type=int, default=None,
                                      help='use PPLACER_CPUS during placement (default: CPUS)')
    optional_classify_wf.add_argument('--force', action='store_const', const=True, default=False,
                                      help='continue processing if an error occurs on a single genome')
    optional_classify_wf.add_argument('--scratch_dir', help='Reduce memory usage by writing to disk (slower).')
    optional_classify_wf.add_argument('-r', '--recalculate_red', action='store_true',
                                      help='recalculate RED values based on the reference tree and all added user genomes')
    # optional_classify_wf.add_argument('-s', '--split_tree', action='store_true',
    #                                   help='Use shards of the reference tree (for Bacteria only). reduce memory usage (slower).')
    optional_classify_wf.add_argument('-d', '--debug', action="store_true",
                                      help='create intermediate files for debugging purposes')
    optional_classify_wf.add_argument('-h', '--help', action="help",
                                      help="show help message")

    # identify marker genes in genomes
    identify_parser = subparsers.add_parser('identify', conflict_handler='resolve',
                                            formatter_class=CustomHelpFormatter,
                                            help='Identify marker genes in genome.')

    mutex_identify = identify_parser.add_argument_group('mutually exclusive required arguments')
    mutex_group = mutex_identify.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument('--genome_dir',
                             help="directory containing genome files in FASTA format")
    mutex_group.add_argument('--batchfile',
                             help="file describing genomes - tab separated in 3 columns (FASTA file, genome ID, translation table [optional])")

    required_identify = identify_parser.add_argument_group('required named arguments')
    required_identify.add_argument('--out_dir', required=True,
                                   help="directory to output files")

    optional_identify = identify_parser.add_argument_group('optional arguments')
    optional_identify.add_argument('-x', '--extension', default='fna',
                                   help='extension of files to process, gz = gzipped')
    optional_identify.add_argument('--prefix', default='gtdbtk',
                                   help='desired prefix for output files')
    optional_identify.add_argument('--cpus', default=1, type=int,
                                   help='number of CPUs to use')
    optional_identify.add_argument('--force', action='store_const', const=True, default=False,
                                   help='continue processing if an error occurs on a single genome')
    optional_identify.add_argument('-h', '--help', action="help",
                                   help="show help message")

    # create multiple sequence alignment
    align_parser = subparsers.add_parser('align', conflict_handler='resolve',
                                         formatter_class=CustomHelpFormatter,
                                         help='Create multiple sequence alignment.', )

    required_align = align_parser.add_argument_group('required named arguments')
    required_align.add_argument('--identify_dir', required=True,
                                help="output directory of 'identify' command")
    required_align.add_argument('--out_dir', required=True,
                                help='directory to output files')

    optional_align = align_parser.add_argument_group('optional arguments')
    optional_align.add_argument('--skip_gtdb_refs',
                                action="store_true",
                                help='do not include GTDB reference genomes in multiple sequence alignment')
    optional_align.add_argument('--taxa_filter',
                                help=('filter GTDB genomes to taxa (comma separated) within '
                                      + 'specific taxonomic groups (e.g., d__Bacteria '
                                      + 'or p__Proteobacteria, p__Actinobacteria)'))
    optional_align.add_argument('--min_perc_aa', type=float, default=10,
                                help='filter genomes with an insufficient percentage of AA in the MSA (inclusive bound)')

    mutual_genome_align = align_parser.add_argument_group('mutually exclusive optional arguments')
    mutex_align_group = mutual_genome_align.add_mutually_exclusive_group()
    mutex_align_group.add_argument('--custom_msa_filters', action="store_true",
                                   help=('perform custom filtering of MSA with cols_per_gene, min_consensus '
                                         + 'max_consensus, and min_perc_taxa parameters instead of using canonical mask'))
    mutex_align_group.add_argument('--skip_trimming', action="store_true", default=False,
                                   help='skip trimming step and return the full MSAs')

    optional_align.add_argument('--cols_per_gene', type=int, default=42,
                                help='maximum number of columns to retain per gene')
    optional_align.add_argument('--min_consensus', type=float, default=25,
                                help='minimum percentage of the same amino acid required to retain column (inclusive bound)')
    optional_align.add_argument('--max_consensus', type=float, default=95,
                                help='maximum percentage of the same amino acid required to retain column (exclusive bound)')
    optional_align.add_argument('--min_perc_taxa', type=float, default=50,
                                help='minimum percentage of taxa required to retain column (inclusive bound)')
    optional_align.add_argument('--rnd_seed', type=int, default=None,
                                help='random seed to use for selecting columns')
    optional_align.add_argument('--prefix', required=False, default='gtdbtk',
                                help='desired prefix for output files')
    optional_align.add_argument('--cpus', default=1, type=int,
                                help='number of CPUs to use')
    optional_align.add_argument('--debug', action="store_true",
                                help='create intermediate files for debugging purposes')
    optional_align.add_argument('-h', '--help', action="help",
                                help="show help message")

    # infer tree
    infer_parser = subparsers.add_parser('infer', conflict_handler='resolve',
                                         formatter_class=CustomHelpFormatter,
                                         help='Infer tree from multiple sequence alignment.', )

    required_infer = infer_parser.add_argument_group('required named arguments')
    required_infer.add_argument('--msa_file', required=True,
                                help="multiple sequence alignment in FASTA format")
    required_infer.add_argument('--out_dir', required=True,
                                help='directory to output files')

    optional_infer = infer_parser.add_argument_group('optional arguments')
    optional_infer.add_argument('--prot_model', choices=['JTT', 'WAG', 'LG'],
                                help='protein substitution model for tree inference', default='WAG')
    optional_infer.add_argument('--no_support', action="store_true",
                                help="do not compute local support values using the Shimodaira-Hasegawa test")
    optional_infer.add_argument('--gamma', action="store_true",
                                help="rescale branch lengths to optimize the Gamma20 likelihood")
    optional_infer.add_argument('--prefix', required=False, default='gtdbtk',
                                help='desired prefix for output files')
    optional_infer.add_argument('--cpus', default=1, type=int,
                                help='number of CPUs to use')
    optional_infer.add_argument('-h', '--help', action="help",
                                help="show help message")

    # classify genomes via placement with pplacer
    classify_parser = subparsers.add_parser('classify', conflict_handler='resolve',
                                            formatter_class=CustomHelpFormatter,
                                            help='Determine taxonomic classification of genomes.', )

    mutual_genome_classify = classify_parser.add_argument_group('mutually exclusive required arguments')
    mutex_group = mutual_genome_classify.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument('--genome_dir',
                             help="directory containing genome files in FASTA format")
    mutex_group.add_argument('--batchfile',
                             help="file describing genomes - tab separated in 2 columns (FASTA file, genome ID)")

    required_classify = classify_parser.add_argument_group('required named arguments')
    required_classify.add_argument('--align_dir', required=True,
                                   help="output directory of 'align' command")
    required_classify.add_argument('--out_dir', required=True,
                                   help='directory to output files')

    optional_classify = classify_parser.add_argument_group('optional arguments')
    optional_classify.add_argument('-x', '--extension', default='fna',
                                   help='extension of files to process, gz = gzipped')
    optional_classify.add_argument('--prefix', required=False, default='gtdbtk',
                                   help='desired prefix for output files')
    optional_classify.add_argument('--cpus', default=1, type=int,
                                   help='number of CPUs to use')
    optional_classify.add_argument('--pplacer_cpus', type=int, default=None,
                                   help='use PPLACER_CPUS during placement (default: CPUS)')
    optional_classify.add_argument('--scratch_dir', help='reduce memory usage by writing to disk (slower)')
    # optional_classify.add_argument('-s', '--split_tree', action='store_true',
    #                                help='Use shards of the reference tree (for Bacteria only). reduce memory usage (slower).')
    optional_classify.add_argument('-r', '--recalculate_red', action='store_true',
                                   help='recalculate RED values based on the reference tree and all added user genomes')

    optional_classify.add_argument('--debug', action="store_true",
                                   help='create intermediate files for debugging purposes')
    optional_classify.add_argument('-h', '--help', action="help",
                                   help="show help message")

    # root tree using outgroup
    root_parser = subparsers.add_parser('root', conflict_handler='resolve',
                                        formatter_class=CustomHelpFormatter,
                                        help='Root tree using an outgroup.', )

    required_root = root_parser.add_argument_group('required named arguments')
    required_root.add_argument('--input_tree', required=True,
                               help="tree to root in Newick format")
    required_root.add_argument('--outgroup_taxon', required=True,
                               help="taxon to use as outgroup (e.g., p__Patescibacteria or p__Altiarchaeota)")
    required_root.add_argument('--output_tree', required=True,
                               help='output tree')

    optional_root = root_parser.add_argument_group('optional arguments')
    optional_root.add_argument('--gtdbtk_classification_file',
                               help="file with GTDB-Tk classifications produced by the `classify` command")
    optional_root.add_argument('--custom_taxonomy_file',
                               help="file indicating custom taxonomy strings for user genomes, which should contain any genomes belonging to the outgroup")
    optional_root.add_argument('-h', '--help', action="help",
                               help="show help message")

    # decorate tree
    decorate_parser = subparsers.add_parser('decorate', conflict_handler='resolve',
                                            formatter_class=CustomHelpFormatter,
                                            help='Decorate tree with GTDB taxonomy.', )

    required_decorate = decorate_parser.add_argument_group('required named arguments')
    required_decorate.add_argument('--input_tree', required=True,
                                   help="tree to root in Newick format")
    required_decorate.add_argument('--output_tree', required=True,
                                   help='output tree')

    optional_decorate = decorate_parser.add_argument_group('optional arguments')
    optional_decorate.add_argument('--gtdbtk_classification_file',
                                   help="file with GTDB-Tk classifications produced by the `classify` command")
    optional_decorate.add_argument('--custom_taxonomy_file',
                                   help="file indicating custom taxonomy strings for user genomes")
    optional_decorate.add_argument('-h', '--help', action="help",
                                   help="show help message")

    # establish taxonomic ranks of internal nodes using RED
    infer_ranks_parser = subparsers.add_parser('infer_ranks', conflict_handler='resolve',
                                               formatter_class=CustomHelpFormatter,
                                               help='Establish taxonomic ranks of internal nodes using RED.', )

    infer_ranks_req = infer_ranks_parser.add_argument_group('required named arguments')
    infer_ranks_req.add_argument('--input_tree', required=True,
                                 help="rooted input tree with labelled ingroup taxon")
    infer_ranks_req.add_argument('--ingroup_taxon', required=True,
                                 help="labelled ingroup taxon to use as root for establish RED values (e.g., c__Bacilli or f__Lactobacillaceae")
    infer_ranks_req.add_argument('--output_tree', required=True,
                                 help="output tree")

    infer_ranks_opt = infer_ranks_parser.add_argument_group('optional arguments')
    infer_ranks_opt.add_argument('-h', '--help', action="help", help="show help message")

    # ani_rep
    ani_rep_parser = subparsers.add_parser('ani_rep', conflict_handler='resolve',
                                           formatter_class=CustomHelpFormatter,
                                           help='Calculates ANI to GTDB representative genomes.', )

    # ani_rep mutex required input genomes
    ani_rep_mutex_genome = ani_rep_parser.add_argument_group('mutually exclusive required arguments')
    ani_rep_mutex_in = ani_rep_mutex_genome.add_mutually_exclusive_group(required=True)
    ani_rep_mutex_in.add_argument('--genome_dir',
                                  help="directory containing genome files in FASTA format")
    ani_rep_mutex_in.add_argument('--batchfile',
                                  help="file describing genomes - tab separated in 2 columns (FASTA file, genome ID)")

    # ani_rep required arguments
    ani_rep_req = ani_rep_parser.add_argument_group('required named arguments')
    ani_rep_req.add_argument('--out_dir', required=True,
                             help="directory to output files")

    # ani_rep mash arguments
    ani_rep_mash = ani_rep_parser.add_argument_group('optional Mash arguments')
    ani_rep_mash.add_argument('--no_mash', action='store_const', const=True, default=False,
                              help='skip pre-filtering using MASH')
    ani_rep_mash.add_argument('--mash_k', default=16, type=int, help='k-mer size [1-32]')
    ani_rep_mash.add_argument('--mash_s', default=5000, type=int, help='maximum number of non-redundant hashes')
    ani_rep_mash.add_argument('--mash_d', default=0.1, type=float, help='maximum distance to keep [0-1]')
    ani_rep_mash.add_argument('--mash_v', default=1.0, type=float, help='maximum p-value to keep [0-1]')

    ani_rep_fastani_opt = ani_rep_parser.add_argument_group('optional FastANI arguments')
    ani_rep_fastani_opt.add_argument('--min_af', default=AF_THRESHOLD, type=float,
                                     help='alignment fraction to consider closest genome')

    # ani_rep optional arguments
    ani_rep_opt = ani_rep_parser.add_argument_group('optional arguments')

    ani_rep_opt.add_argument('-x', '--extension', default='fna',
                             help='extension of files to process, gz = gzipped')
    ani_rep_opt.add_argument('--prefix', default='gtdbtk',
                             help='desired prefix for output files')
    ani_rep_opt.add_argument('--cpus', default=1, type=int, help='number of CPUs to use')
    ani_rep_opt.add_argument('-h', '--help', action="help", help="show help message")

    # test
    test_parser = subparsers.add_parser('test', conflict_handler='resolve',
                                        formatter_class=CustomHelpFormatter,
                                        help='Test the classify_wf pipeline with 3 archaeal genomes.')
    required_test = test_parser.add_argument_group('required named arguments')
    required_test.add_argument('--out_dir', required=True,
                               help='directory to output files')
    optional_test = test_parser.add_argument_group('optional arguments')
    optional_test.add_argument('--cpus', default=1, type=int,
                               help='number of CPUs to use')
    optional_test.add_argument('-h', '--help', action="help",
                               help="show help message")

    # trim MSA
    msa_parser = subparsers.add_parser('trim_msa', conflict_handler='resolve',
                                       formatter_class=CustomHelpFormatter,
                                       help='Trim an untrimmed MSA file based on a mask.', )

    required_msa = msa_parser.add_argument_group('required named arguments')
    required_msa.add_argument('--untrimmed_msa', required=True,
                              help="untrimmed MSA file")
    required_msa.add_argument('--output', required=True,
                              help='output file')
    mutual_trim_msa = msa_parser.add_argument_group('mutually exclusive required arguments')
    mutex_msa_group = mutual_trim_msa.add_mutually_exclusive_group(required=True)
    mutex_msa_group.add_argument('--mask_file',
                                 help="mask file to use for trimming the MSA")
    mutex_msa_group.add_argument('--reference_mask',
                                 choices=['arc', 'bac'],
                                 help="reference mask already present in GTDB-Tk")

    optional_msa = msa_parser.add_argument_group('optional arguments')
    optional_msa.add_argument('-h', '--help', action="help",
                              help="show help message")

    # export msa
    export_msa_parser = subparsers.add_parser('export_msa', conflict_handler='resolve',
                                              formatter_class=CustomHelpFormatter,
                                              help='Export the untrimmed archaeal or bacterial MSA file.', )
    required_export_msa = export_msa_parser.add_argument_group('required named arguments')
    required_export_msa.add_argument('--domain', required=True, choices=['arc', 'bac'],
                                     help="select domain to download")
    required_export_msa.add_argument('--output', required=True,
                                     help='output file')
    optional_export_msa = export_msa_parser.add_argument_group('optional arguments')
    optional_export_msa.add_argument('-h', '--help', action="help",
                                     help="show help message")

    # verify install
    check_install_parser = subparsers.add_parser('check_install', conflict_handler='resolve',
                                                 formatter_class=CustomHelpFormatter,
                                                 help='Verify if all gtdb data files are present to run GTDB-Tk.', )
    optional_check_install = check_install_parser.add_argument_group('optional arguments')
    optional_check_install.add_argument('-h', '--help', action="help",
                                        help="show help message")

    # -------------------------------------------------
    # get and check options
    args = None
    if len(sys.argv) == 1:
        print_help()
        sys.exit(0)
    elif sys.argv[1] in {'-v', '--v', '-version', '--version'}:
        print("gtdbtk: version %s %s %s" % (__version__,
                                            __copyright__,
                                            __author__))
        sys.exit(0)
    elif sys.argv[1] in {'-h', '--h', '-help', '--help'}:
        print_help()
        sys.exit(0)
    else:
        args = parser.parse_args()

    # setup logger
    logger_setup(args.out_dir if hasattr(args, 'out_dir') else None,
                 "gtdbtk.log", "GTDB-Tk", __version__, False,
                 hasattr(args, 'debug') and args.debug)
    logger = logging.getLogger('timestamp')

    # -------------------------------------------------
    # do what we came here to do
    try:
        gt_parser = OptionsParser(__version__)
        if False:
            import cProfile

            cProfile.run('gt_parser.parseOptions(args)', 'prof')
        else:
            gt_parser.parse_options(args)
    except SystemExit:
        sys.stdout.write('\n')
        sys.stdout.flush()
        logger.error('Controlled exit resulting from early termination.')
        sys.exit(1)
    except KeyboardInterrupt:
        sys.stdout.write('\n')
        sys.stdout.flush()
        logger.error('Controlled exit resulting from interrupt signal.')
        sys.exit(1)
    except GTDBTkExit as e:
        sys.stdout.write('\n')
        sys.stdout.flush()
        if len(str(e)) > 0:
            logger.error('{}'.format(e))
        logger.error('Controlled exit resulting from an unrecoverable error or warning.')
        sys.exit(1)
    except (GTDBTkException, BioLibError) as e:
        sys.stdout.write('\n')
        sys.stdout.flush()
        msg = 'Controlled exit resulting from an unrecoverable error or warning.\n\n'
        msg += '=' * 80 + '\n'
        msg += 'EXCEPTION: {}\n'.format(type(e).__name__)
        msg += '  MESSAGE: {}\n'.format(e)
        msg += '_' * 80 + '\n\n'
        msg += traceback.format_exc()
        msg += '=' * 80
        logger.error(msg)
        sys.exit(1)
    except Exception as e:
        sys.stdout.write('\n')
        sys.stdout.flush()
        msg = 'Uncontrolled exit resulting from an unexpected error.\n\n'
        msg += '=' * 80 + '\n'
        msg += 'EXCEPTION: {}\n'.format(type(e).__name__)
        msg += '  MESSAGE: {}\n'.format(e)
        msg += '_' * 80 + '\n\n'
        msg += traceback.format_exc()
        msg += '=' * 80
        logger.error(msg)
        sys.exit(1)


if __name__ == '__main__':
    main()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging

import dendropy

from gtdbtk.biolib_lite.newick import parse_label, create_label
from gtdbtk.biolib_lite.taxonomy import Taxonomy
from gtdbtk.config.config import (TAXONOMY_FILE,
                                  RED_DIST_BAC_DICT,
                                  RED_DIST_ARC_DICT,
                                  MRCA_RED_BAC120,
                                  MRCA_RED_AR122,
                                  RED_INTERVAL)
from gtdbtk.exceptions import GTDBTkExit
from gtdbtk.relative_distance import RelativeDistance


class InferRanks(object):
    """Establish taxonomic ranks of internal nodes using RED."""

    def __init__(self):
        """Initialize."""

        self.logger = logging.getLogger('timestamp')

    def _get_ingroup_domain(self, ingroup_taxon) -> str:
        """Get domain on ingroup taxon."""

        # read GTDB taxonomy in order to establish domain on ingroup taxon
        gtdb_taxonomy = Taxonomy().read(TAXONOMY_FILE)
        ingroup_domain = None
        for taxa in gtdb_taxonomy.values():
            if ingroup_taxon in taxa:
                ingroup_domain = taxa[Taxonomy.DOMAIN_IDX]

        if ingroup_domain is None:
            raise GTDBTkExit(f'Ingroup taxon {ingroup_taxon} was not found in '
                             f'the GTDB taxonomy.')

        return ingroup_domain

    def _get_median_reds(self, ingroup_domain: str):
        """Get median RED values for domain of ingroup taxon."""

        # get median RED values for domain
        if ingroup_domain == 'd__Bacteria':
            median_reds = RED_DIST_BAC_DICT
        elif ingroup_domain == 'd__Archaea':
            median_reds = RED_DIST_ARC_DICT
        else:
            raise GTDBTkExit(f'Unrecognized GTDB domain: {ingroup_domain}.')

        # report median values
        domain = ingroup_domain.replace('d__', '')
        self.logger.info('Median RED values for {}:'.format(domain))
        for idx, rank_prefix in enumerate(Taxonomy.rank_prefixes):
            if idx != Taxonomy.DOMAIN_IDX and idx != Taxonomy.SPECIES_IDX:
                self.logger.info('  {}\t{:.3f}'.format(
                    Taxonomy.rank_labels[idx].capitalize(),
                    median_reds[rank_prefix]))

        return median_reds

    def _find_ingroup_taxon(self, ingroup_taxon, tree):
        """Find node of ingroup taxon in tree."""

        ingroup_node = None
        for node in tree.postorder_node_iter():
            support, taxon, auxiliary_info = parse_label(node.label)

            if taxon:
                taxa = [t.strip() for t in taxon.split(';')]
                if ingroup_taxon in taxa:
                    if ingroup_node is not None:
                        raise GTDBTkExit(f'Ingroup taxon {ingroup_taxon} '
                                         f'identified multiple times.')
                    ingroup_node = node

        if ingroup_node is None:
            raise GTDBTkExit(f'Ingroup taxon {ingroup_taxon} not found in tree.')

        return ingroup_node

    def _find_ingroup_red(self, ingroup_node, ingroup_domain, tree):
        """Find RED of the ingroup taxon."""

        red_file = MRCA_RED_BAC120
        if ingroup_domain == 'd__Archaea':
            red_file = MRCA_RED_AR122

        # create map from leave labels to tree nodes
        leaf_node_map = {}
        for leaf in tree.leaf_node_iter():
            leaf_node_map[leaf.taxon.label] = leaf

        # find RED value of ingroup node
        reference_nodes = set()
        with open(red_file) as rf:
            for line in rf:
                label_ids, red = line.strip().split('\t')
                labels = label_ids.split('|')
                if len(labels) == 2:
                    taxa = [leaf_node_map[label].taxon for label in labels]
                    node = tree.mrca(taxa=taxa)
                    if node == ingroup_node:
                        return float(red)

        raise GTDBTkExit(f'Could not determine RED of ingroup taxon {ingroup_node}.')

    def _determine_red_ranks(self, node_red, median_reds):
        """Determine suitable taxonomic ranks for node using RED."""

        red_ranks = {}
        for rank_prefix, median_red in median_reds.items():
            rank_idx = Taxonomy.rank_index[rank_prefix]
            rank_label = Taxonomy.rank_labels[rank_idx]

            abs_red_diff = abs(node_red - median_red)
            if abs_red_diff <= RED_INTERVAL:
                red_ranks[rank_label] = abs_red_diff

        red_ranks_label = []
        for rank_label, abs_red_diff in sorted(red_ranks.items(), key=lambda kv: kv[1]):
            red_ranks_label.append(rank_label)

        return '&'.join(red_ranks_label)

    def run(self, input_tree, ingroup_taxon, output_tree):
        """Establish taxonomic ranks of internal nodes using RED..

        Parameters
        ----------
        input_tree : str
          Rooted tree with labelled outgroup.
        ingroup_taxon : str
          Ingroup from which to infer ranks based on RED.
        output_tree: str
          Output directory.
        """

        # get domain on ingroup taxon
        ingroup_domain = self._get_ingroup_domain(ingroup_taxon)

        # get median RED values for domain of ingroup taxon
        median_reds = self._get_median_reds(ingroup_domain)

        # read tree
        self.logger.info('Reading tree.')
        tree = dendropy.Tree.get_from_path(input_tree,
                                           schema='newick',
                                           rooting='force-rooted',
                                           preserve_underscores=True)

        # find ingroup taxon
        ingroup_node = self._find_ingroup_taxon(ingroup_taxon, tree)

        # get RED of ingroup taxon
        ingroup_red = self._find_ingroup_red(ingroup_node, ingroup_domain, tree)
        self.logger.info('RED of ingroup taxon {} = {:.3f}'.format(
            ingroup_taxon, ingroup_red))

        # get RED value of ingroup taxon
        self.logger.info('Decorating tree with RED and rank information.')
        red = RelativeDistance()
        red.decorate_rel_dist(ingroup_node, ingroup_red)

        for node in ingroup_node.preorder_iter():
            if node.is_leaf():
                continue

            support, taxon, auxiliary_info = parse_label(node.label)

            if auxiliary_info:
                auxiliary_info += '|RED={:.3f}'.format(node.rel_dist)
            else:
                auxiliary_info = 'RED={:.3f}'.format(node.rel_dist)

            red_ranks = self._determine_red_ranks(node.rel_dist, median_reds)
            auxiliary_info += '|{}'.format(red_ranks)

            new_label = create_label(support, taxon, auxiliary_info)
            node.label = new_label

        # write RED decorated tree to file
        tree.write_to_path(output_tree,
                           schema='newick',
                           suppress_rooting=True,
                           unquoted_underscores=True)
